@@ -203,6 +203,56 @@ def fsck(device):
         sys.exit(1)
 
 
+def mkfs(device, fs_type, label):
+    '''
+    Run mkfs(8) on specified device.
+    '''
+    assert (os.path.exists(device))
+
+    args = ['/sbin/mkfs',
+            '-v'
+            '-t', fs_type,
+            '-L', label,
+            device]
+
+    log.debug('running: {}'.format(args))
+
+    proc = subprocess.Popen(args,
+                            stdout=subprocess.PIPE,
+                            stderr=subprocess.PIPE,
+                            universal_newlines=True)
+
+    ret = proc.wait()
+
+    if ret == 0:
+        return
+
+    stdout, stderr = proc.communicate()
+
+    log.error('{} returned {}: {}, {}'
+              .format(args,
+                      ret,
+                      stdout,
+                      stderr))
+    sys.exit(1)
+
+
+def get_mount_details(target):
+    '''
+    Returns a list comprising device, filesystem type and filesystem
+    label for the mount target specified.
+    '''
+    args = ['findmnt', '-o', 'SOURCE,FSTYPE,LABEL', '-n', target]
+    output = subprocess.check_output(args, universal_newlines=True)
+    output = output.strip().split()
+
+    if len(output) != 3:
+        sys.exit('Failed to determine mount details for {}'
+                 .format(target))
+
+    return output
+
+
 def remount(mountpoint, options):
     """
     Remount mountpoint using the specified options string (which is
@@ -468,33 +518,31 @@ class Upgrader():
         # If True, the other partition is considered empty. In this
         # scenario, prior to unpacking the latest image, the _current_
         # rootfs image is copied to the other partition.
-        #
-        # Note: Only used by UPGRADE_IN_PLACE.
         self.other_is_empty = False
 
+        # True if the upgrader has reformatted the "other" partition.
+        self.other_has_been_formatted = False
+
+        # cache_dir is used as a scratch pad, for downloading new images
+        # to and bind mounting the rootfs.
+        if self.options.tmpdir:
+            self.cache_dir = self.options.tmpdir
+        else:
+            self.cache_dir = self.DEFAULT_CACHE_DIR
+
     def update_timestamp(self):
         '''
         Update the timestamp file to record the time the last upgrade
         completed successfully.
         '''
-        file = os.path.join(self.get_cache_dir(), self.TIMESTAMP_FILE)
+        file = os.path.join(self.cache_dir, self.TIMESTAMP_FILE)
         open(file, 'w').close()
 
-    def get_cache_dir(self):
-        '''
-        Returns the full path to the cache directory, which is used as a
-        scratch pad, for downloading new images to and bind mounting the
-        rootfs.
-        '''
-        return self.options.tmpdir \
-            if self.options.tmpdir \
-            else self.DEFAULT_CACHE_DIR
-
     def get_mount_target(self):
         '''
         Get the full path to the mount target directory.
         '''
-        return os.path.join(self.get_cache_dir(), self.MOUNT_TARGET)
+        return os.path.join(self.cache_dir, self.MOUNT_TARGET)
 
     def prepare(self):
         '''
@@ -566,9 +614,17 @@ class Upgrader():
         if self.options.dry_run or self.options.root_dir != '/':
             return
 
-        log.warning('ingoring format target {} (unsupported operation)'
-                    .format(target))
-        return
+        other = self.get_mount_target()
+
+        source, fstype, label = get_mount_details(other)
+
+        unmount(other)
+        mkfs(source, fstype, label)
+
+        # leave the mount as it was initially.
+        mount(source, other, "ro")
+
+        self.other_has_been_formatted = True
 
     def _cmd_load_keyring(self, args):
         try:
@@ -609,6 +665,19 @@ class Upgrader():
                         .format(target_type))
             return
 
+        if self.other_is_empty and not self.other_has_been_formatted:
+            # We believe "other" is empty. However, it's possible that a
+            # previous attempt at upgrading failed due to a power
+            # outage. In that scenario, "other" may contain most of a
+            # rootfs image, but just be missing the files used to
+            # determine that the partition is empty. As such, if we
+            # believe the partition is empty, it should forcibly be made
+            # empty since it may contain detritus from a
+            # previously-failed unpack (possibly caused by a power
+            # outage).
+            log.warning('reformatting other (no system image)')
+            self._cmd_format("system")
+
         self.remount_rootfs(writable=True)
 
         if self.other_is_empty:
@@ -739,14 +808,19 @@ class Upgrader():
         else:
             to_remove = []
 
+        # by definition, full images don't have 'removed' files.
         if not self.full_image:
-
             if found_removed_file:
                 log.debug('processing {} file'
                           .format(self.removed_file))
 
                 # process backwards to work around bug LP:#1381134.
                 for remove in sorted(to_remove, reverse=True):
+
+                    if not remove:
+                        # handle invalid removed file entry (see LP:#1437225)
+                        continue
+
                     remove = remove.strip()
 
                     # don't remove devices
@@ -762,13 +836,7 @@ class Upgrader():
                     if '../' in remove:
                         continue
 
-                    if self.options.root_dir == '/':
-                        final = os.path.join(self.get_cache_dir(), remove)
-                    else:
-                        # os.path.join() refuses to work if the file
-                        # begins with a slash.
-                        base = remove_prefix(remove)
-                        final = '{}{}'.format(self.options.root_dir, base)
+                    final = os.path.join(self.cache_dir, remove)
 
                     if not os.path.exists(final):
                         # This scenario can only mean there is a bug
@@ -804,9 +872,9 @@ class Upgrader():
             from unittest.mock import patch
             with patch("grp.getgrnam") as m:
                 m.side_effect = KeyError()
-                tar.extractall(path=self.get_cache_dir(),
+                tar.extractall(path=self.cache_dir,
                                members=tar_generator(
-                                   tar, self.get_cache_dir(),
+                                   tar, self.cache_dir,
                                    self.removed_file,
                                    self.options.root_dir))
         tar.close()
@@ -816,12 +884,12 @@ class Upgrader():
         Copy all rootfs data from the current partition to the other
         partition.
 
-        XXX: Assumes that the other rootfs is already mounted to
+        XXX: Assumes that the other rootfs is already mounted r/w to
         mountpoint get_mount_target().
         '''
         target = self.get_mount_target()
         bindmount_rootfs_dir = tempfile.mkdtemp(prefix=script_name,
-                                                dir=self.get_cache_dir())
+                                                dir=self.cache_dir)
         bind_mount("/", bindmount_rootfs_dir)
 
         cwd = os.getcwd()
@@ -846,4 +914,3 @@ class Upgrader():
 
         unmount(bindmount_rootfs_dir)
         os.rmdir(bindmount_rootfs_dir)
-        self.other_is_empty = False
